

def statement_for_total_ooh(df_asset_group_period_sum_others, row_index_for_total='Total', period_freq_str='week'):

    sub_pct_value = df_asset_group_period_sum_others['sub_pct'][row_index_for_total]
    sub_pct_abs_value = round(abs(sub_pct_value * 100))

    if sub_pct_abs_value > 1:
        if sub_pct_value > 0:
            statement_direction = "up"
        else:
            statement_direction = "down"
        statement = f"""The out-of-hours use has gone {statement_direction} by {sub_pct_abs_value}% compared to the {period_freq_str} before.""" #temp fix

    else:   
        statement = f"""The out-of-hours use has been similar to the {period_freq_str} before.""" #temp fix
        
    return statement



import matplotlib.ticker as ticker

def setup_y_tick_lim_label_zero_line(ax, ls_ylim, display_unit, formatter,
                                     fontsize=None):

    if fontsize is None:
        fontsize = 13

    # move the y_tick to the right
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")
    ax.tick_params(axis='y', right=True, length=5, pad=9, labelsize=fontsize)
    ax.set_ylim(ls_ylim)

    # set y label
    ax.set_ylabel(f"Daily Usage ({display_unit})", labelpad=18, fontsize=fontsize)

    # set yaxis formatter
    ax.yaxis.set_major_formatter(ticker.FormatStrFormatter(formatter))

    # plot the zero value horizontal line
    ax.axhline(y=0, color='k', linestyle=':', zorder=-1)


import smtplib

def get_smtp_server(server, port):

    smtp_server = smtplib.SMTP(server, port)

    # the below part is to make the code compatible in the synapse environment
    # by changing the local_hostname to all lowercase
    (code, resp) = smtp_server.helo()
    if code == 501:
        # parse the local_hostname
        localhost_with_time = str(resp).split('[')[1]
        localhost_name = localhost_with_time.split(' ')[0]

        # apply the lowercase local hostname
        localhost_in_lowercase = localhost_name.lower()
        smtp_server = smtplib.SMTP(server, port, local_hostname=localhost_in_lowercase)
        # server.ehlo(localhost_in_lowercase)
    
    return smtp_server


from pathlib import Path
import matplotlib.pyplot as plt
from datetime import date
import io

from .preprocessing_for_barchart import preprocessing_for_barchart
from .plot_shifted_daily_barchart_with_temp import plot_shifted_daily_barchart_with_temp

def generate_daily_barchart_with_temp(df_meta_with_value, 
                                      sr_temp=None,
                                      fontsize=None, # please note that the other fontsizes haven't been tested yet
                                      directory_to_savefig=None):

    df_pivot_working_hours_sorted, period_current = preprocessing_for_barchart(df_meta_with_value)

    # df_pivot_working_hours_sorted[(True, True)] -= [0.2, 0.1, 0.05, 0.15, 0.08, 0.12, 0.21]
    # df_pivot_working_hours_sorted[(True, False)] -= [0.12, 0.16, 0.08, 0.12, 0.21, 0.05, 0.15]
    # df_pivot_working_hours_sorted[True] = df_pivot_working_hours_sorted[True] *0.5
    # df_pivot_working_hours_sorted[True] -= 0.4*df_pivot_working_hours_sorted[False]

    # df_pivot_working_hours_sorted_reset = df_pivot_working_hours_sorted.reset_index(drop=True)
    # sr_temp = df_pivot_working_hours_sorted_reset[True].sum(axis=1).abs()

    
    plot_shifted_daily_barchart_with_temp(df_pivot_working_hours_sorted,
                                          sr_temp=sr_temp,
                                          fontsize=fontsize)

    png_name = 'daily_barchart_with_temp.png'
    
    if directory_to_savefig == None:
        png_object = io.BytesIO()
        plt.savefig(png_object, format='png', dpi=200)
        plt.close()
        return {png_name: png_object}
    else:
        png_path = directory_to_savefig + png_name
        plt.savefig(png_path, format='png', dpi=200)
        plt.close()
        return {png_name: png_path}




import pandas as pd
from MondayProcessor import MondayProcessor

def get_last_update_time(monday_token, board_id):

    mp = MondayProcessor(monday_token)

    js_logs = mp._fetchBoardLogsById(board_id)
    
    dct_records = js_logs['data']['boards'][0]['activity_logs']

    # print('dct_records: ', dct_records)

    if dct_records == None: # it happened on 11 Nov 24 that the activity log is not available and returned None

        print('js_logs: ', js_logs)
        print('[WARN]: using time now as last_update_time because the Monday ACTIVITY LOG is NOT AVAILABLE !')
        last_update_time = pd.Timestamp.now(tz='UTC').ceil(freq='s')
        return last_update_time
    else:
        df_logs = pd.DataFrame.from_records(dct_records)
        df_logs['created_at'] = pd.to_datetime(df_logs['created_at'].astype('int').div(10), unit='us', utc=True).dt.ceil(freq='s') 
        last_update_time = df_logs.created_at.max()

        return last_update_time



import pandas as pd

from .get_meta_statement import get_meta_statement

def load_metadata(engine, table_name_meta, building_id, 
                #   organisation=None,  
                  exception=None):

    statement_full = get_meta_statement(building_id, 
                    #    organisation=organisation, 
                       exception=exception, # the filter will be applied to link type
                      )
    
    # print('statement_full: ', statement_full)

    meta_query = f"""select id, 
                            org,
                            building_name,
                            thing_name, 
                            thing_type, 
                            thing_category,
                            variable,
                            phase,
                            nid,
                            channel,
                            hardware_id
                            from {table_name_meta} 
                            where {statement_full};"""

    with engine.connect() as conn:
        df_meta = pd.read_sql_query(meta_query, con=conn)

    return df_meta


from typing import List

from .MondayProcessorExceptions_ import (
    MondayProcessorBoardNotFound,
    MondayProcessorNoColumInBoard
)

def _fetchGroups(self, board_id:int) -> List[str]:
    data = self.monday.groups.get_groups_by_board(board_ids=board_id)
    boards = data['data']['boards']
    if(len(boards) <1):
        raise MondayProcessorBoardNotFound(f"Monday board id ({board_id}) not found when perform MondayProcessor._searchConfColumnIds")
    groups = boards[0]['groups']
    if(len(groups) < 1):
        raise MondayProcessorNoColumInBoard(f"No column in board {board_id}")
    ret = {x["id"]:x["title"] for x in groups}
#         print(groups)
    # print(ret)
    return ret


import pandas as pd

def get_period_range(self):

    reference_time = pd.Timestamp.now(tz=self.timezone)  
    period_range = pd.period_range(end=reference_time, freq=self.period_freq, periods=self.period_count+self.period_offset)

    return period_range


from pathlib import Path
import matplotlib.pyplot as plt
import io

from .preprocessing_for_energy_meter_with_benchmarking import preprocessing_for_energy_meter_with_benchmarking
from .energy_meter_with_benchmarking import energy_meter_with_benchmarking


def generate_energy_meter_with_benchmarking(df_meta_with_value_building, floor_size,
                                            industry = "office",
                                            size_in_sqm = True,
                                            conv_mwh_price = 190,
                                            currency = 'GBP',
                                            fontsize = 13,
                                            directory_to_savefig=None):
                                            
    consumption_mwh_cur, consumption_mwh_pre, days_in_period = preprocessing_for_energy_meter_with_benchmarking(df_meta_with_value_building)

    print('consumption_mwh_building_total: ', consumption_mwh_cur)  # debug
    energy_meter_with_benchmarking(consumption_mwh_cur, consumption_mwh_pre, floor_size, 
                                    industry=industry, 
                                    period=days_in_period, 
                                    size_in_sqm=size_in_sqm,
                                    conv_mwh_price=conv_mwh_price,
                                    currency=currency,
                                    fontsize=fontsize)
    
    png_name = 'monthly_total_and_bm_latest.png'
    if directory_to_savefig == None:
        png_object = io.BytesIO()
        plt.savefig(png_object, format='png', dpi=200,transparent=True, bbox_inches='tight', pad_inches=0)
        plt.close()
        return {png_name: png_object}
    else:
        png_path = directory_to_savefig + png_name
        plt.savefig(png_path, format='png', dpi=200,transparent=True, bbox_inches='tight', pad_inches=0)
        plt.close()
        return {png_name: png_path}


    # Path(directory_to_savefig).mkdir(parents=True, exist_ok=True)
    # plt.savefig(directory_to_savefig+"monthly_total_and_bm_latest.png", format='png', dpi=200,transparent=True, bbox_inches='tight', pad_inches=0)




import os

def delete_files_in_directory(directory_path, ignore=[]):
   try:
     files = os.listdir(directory_path)
     filtered_files = C = list(set(files) - set(ignore))
     for file in filtered_files:
       file_path = os.path.join(directory_path, file)
       if os.path.isfile(file_path):
         os.remove(file_path)
     print(f"\nAll files in path [{directory_path}] were deleted successfully.")
   except OSError:
     print("Error occurred while deleting files.")



import matplotlib.pyplot as plt
import copy

from .adjust_gap_for_visibility import adjust_gap_for_visibility
from .adjust_to_inner_ring import adjust_to_inner_ring
from .annotate_pct_on_pies import annotate_pct_on_pies
from .generate_pct_fmt import generate_pct_fmt
from .adjust_values_for_better_visibility import adjust_values_for_better_visibility


def plot_ring(sr_orig, ax,
                colors=None,
                hatch=None,
                fontsize=None,
                as_inner_ring=None,
                inner_radius=None,
                outer_radius=None,
                gap_value=None, 
                ann_pct=None,
                pct_out=None, 
                pct_hide=None):
    
    if fontsize is None:
        fontsize = 13
    if colors is None:
        colors = ['#fffffc'] # use this modification of white color to differentiate it from the actual white, for legend handle filter
    if as_inner_ring is None:
        as_inner_ring = False
    if inner_radius is None:
        inner_radius = 0.5
    if outer_radius is None:
        outer_radius = 1
    if ann_pct is None:
        ann_pct = True
    if gap_value is None:
        gap_value = 0
    
    sr = adjust_values_for_better_visibility(sr_orig) # to avoid changes made to the original series

    ls_orig_pct_fmt = generate_pct_fmt(sr)
    
    if gap_value > 0:

        sr_sum_without_gap = sr.sum()

        # to make the non-gap pie or gap pie more visible
        sr['Gap Value'] = adjust_gap_for_visibility(gap_value, sr_sum_without_gap, 
                                                    min_vis_pct=3)
        
    sr.plot.pie(ax=ax, 
                autopct='', # needed to generate placeholder for ax.texts
                colors=colors,
                hatch=hatch,
                radius=outer_radius,
                # frame=True,
                textprops={"color": 'k', 
                            "fontsize": fontsize}, 
                pctdistance=0.77,
                wedgeprops={'linewidth': 1, 
                            "edgecolor": 'k'}, 
                startangle=90,
                labels=None)
    
    if gap_value > 0:
        patch_for_gap = copy.copy(ax.patches[-1])
        ax.patches[-1].set(visible=False, gid='no legend')

    if ann_pct:
        annotate_pct_on_pies(ax, sr, ls_orig_pct_fmt, 
                            pct_hide=pct_hide,
                            pct_out=pct_out)
    
    # plot the circle in the centre of the ring
    inner_circle = plt.Circle((0, 0), inner_radius, 
                              facecolor='w', 
                              linewidth=1, 
                              edgecolor='k') # edgecolor=other_colours[0], 
    # p = plt.gcf()
    # p.gca().add_artist(inner_circle)
    ax.add_artist(inner_circle)

    if gap_value > 0:
        adjust_to_inner_ring(ax, patch_for_gap, inner_radius, outer_radius, 
                        as_inner_ring=as_inner_ring)

    ax.set_ylabel("")





def write_multicell_with_styles(self, max_width, cell_height, text_list):

        # Source:https://stackoverflow.com/questions/60736940/how-to-make-inline-bold-text-using-pyfpdf#
        startx = self.get_x()
        self.set_font('Arial', '', 12)

        #loop through differenct sections in different styles
        for text_part in text_list:
            #check and set style
            try:
                current_style = text_part['style']
                self.set_font('Arial', current_style, 12)
            except KeyError:
                self.set_font('Arial', '', 12)

            #loop through words and write them down
            space_width = self.get_string_width(' ')
            for word in text_part['text'].split(' '):
                current_pos = self.get_x()
                word_width = self.get_string_width(word)
                #check for newline
                if (current_pos + word_width) > (startx + max_width):
                    #return 
                    self.set_y(self.get_y() + cell_height)
                    self.set_x(startx)
                self.cell(word_width, 5, word)
                #add a space
                self.set_x(self.get_x() + space_width)


# email template setting, note: all of the available variables for email subject and message templates can be accessed from context object ct

# monday_board_id = '6323575451'

reviewer_emails = ['x.yang@cloudfmgroup.com', 
                  #  's.goud@cloudfmgroup.com',
                    #    'a.green@cloudfmgroup.com',
                    #    'a.grimmett@cloudfmgroup.com',
                    #    's.blank@cloudfmgroup.com',
                    #    's.ruthven@cloudfmgroup.com'
                       ]

subject= """Energy Consumption Report - {{ ct['current_period_str'] }}"""
# message= """ Hi {{ ct['receiver_name'] }},

# Please see the energy report for {{ ct['current_period_str'] }}.

# Kind regards,

# Cloudfm Mindsett Team

# """

message = """
<html>
<body>
<p>Hi {{ ct['receiver_name'] }},<br></p>
<p>Please see the energy report for {{ ct['current_period_str'] }}.</p>
<p>Kind regards,</p>
<p>Cloudfm Mindsett Team</p>
</body>
</html>
"""








from mdl.utils_ import modify_group_names
from .exchange_group_name_column import exchange_group_name_column

def refine_group_name(df_meta_with_value, 
                      asset_group, 
                      group_name_column_exchange, 
                      fillna_value,
                      group_name_modification):
    
    df_meta_with_value = exchange_group_name_column(df_meta_with_value, 
                                                    group_name_column_exchange, 
                                                    asset_group)
    
    df_meta_with_value[asset_group] = df_meta_with_value[asset_group].fillna(fillna_value) 

    df_meta_with_value = modify_group_names(df_meta_with_value, 
                                            group_name_column=asset_group,
                                            modification_dict=group_name_modification) 
    return df_meta_with_value


import pandas as pd
from itertools import product

from .patch_missing_ooh_periods_sign import patch_missing_ooh_periods_sign

def preprocessing_for_period_barchart(df_meta_with_value,
                                      sign_column=None,
                                        period_column=None, 
                                        ooh_column=None, 
                                        kwh_column=None,
                                        no_of_period=None):
    if sign_column is None:
        sign_column = 'neg_sign'

    if period_column is None:
        period_column = "period"
    
    if ooh_column is None:
        ooh_column = "out_of_hours"

    if  kwh_column is None:
        kwh_column = "kwh"

    if no_of_period is None:
        no_of_period = 6

    df_meta_with_value[sign_column] = df_meta_with_value[kwh_column].lt(0)

    df_grouped_working_hours_period = df_meta_with_value.groupby([sign_column, period_column, ooh_column])[kwh_column].sum()

    period_max = df_meta_with_value[period_column].max() # todo: check whether there are other better ways for getting the max period, as it might be missing also

    # handle the situation that not all groups exist
    df_grouped_working_hours_period = patch_missing_ooh_periods_sign(df_grouped_working_hours_period, period_max,
                                                                    no_of_period=no_of_period)

    df_grouped_working_hours_period_unstacked = df_grouped_working_hours_period.unstack([sign_column, ooh_column])
    df_grouped_working_hours_period_unstacked = df_grouped_working_hours_period_unstacked.sort_index()

    # convert the unit from kwh to mwh
    df_grouped_working_hours_period_unstacked = df_grouped_working_hours_period_unstacked.div(1000)
    
    return df_grouped_working_hours_period_unstacked


import pandas as pd

def query_building_total(db, start_time, end_time, building_id, 
                        #  organisation=None,
                         timezone='UTC'):

    query_start_time = pd.Timestamp(start_time, tz=timezone)
    query_end_time  =  pd.Timestamp(end_time, tz=timezone)

    query_start_time_str = query_start_time.strftime("%Y-%m-%d %H:%M:%S.%f%z")
    query_end_time_str   = query_end_time.strftime("%Y-%m-%d %H:%M:%S.%f%z")

    # print('query_start_time_str: ', query_start_time_str)
    # print('query_end_time_str: ', query_end_time_str)


    time_period = f"""(time >= '{query_start_time_str}') and (time < '{query_end_time_str}')"""

    statement_list = [f"""building_id = '{building_id}'"""]
    
    # if organisation != None:
    #     statement_new = f"""btrim("organisation") = '{organisation}'"""
    #     statement_list.append(statement_new)

    statement_full = ' and '.join(statement_list)

    query = f"""select time, kwh from {db.table_building_total.name} where {statement_full} and {time_period};"""

    # year, month, building_name, organisation, building_id, org_id

    # print('building_total query: ', query)

    engine = db.engine 

    with engine.connect() as conn:
    
        df_meta_with_value = pd.read_sql_query(query, con=conn)


    return df_meta_with_value



import pandas as pd

def get_tick_values(phase_values_list,
                    min_visible_pct=None):

    if min_visible_pct is None:
        min_visible_pct = 0.077

    df_phase = pd.DataFrame({'phase_value': phase_values_list})

    total_value = df_phase['phase_value'].sum()

    # setup the threshold
    min_visible_value = total_value * min_visible_pct

    # label the visibility
    df_phase['min_visible_value'] = min_visible_value
    df_phase['visible'] = df_phase['phase_value'].gt(min_visible_value)

    # separate by visibility
    df_phase_visible = df_phase.loc[df_phase['visible']]
    df_phase_invisible = df_phase.loc[~df_phase['visible']]

    # get adjusted value for invisible part
    df_phase_invisible['adj_value'] = df_phase_invisible['min_visible_value']

    # get adjusted value for visible part
    compensation = df_phase_invisible['adj_value'].sum() - df_phase_invisible['phase_value'].sum()
    visible_orig_sum = df_phase_visible['phase_value'].sum()
    shrink_pct = (visible_orig_sum - compensation)/visible_orig_sum
    df_phase_visible['adj_value'] = df_phase_visible['phase_value'] * shrink_pct

    # combine the visible and invisible parts back
    df_phase_adj = pd.concat([df_phase_visible, df_phase_invisible]).sort_index()

    # get the tick values
    df_phase_adj['tick_value'] = df_phase_adj['adj_value'].cumsum()

    non_zero_tick_values = df_phase_adj['tick_value'].to_list()

    tick_values_list = [0, *non_zero_tick_values]

    return tick_values_list



from .statement_for_biggest_ooh import statement_for_biggest_ooh
from .statement_for_total_ooh import statement_for_total_ooh
from .statement_for_avg_action_time import statement_for_avg_action_time


def insight_statements(db, df_for_statements, df_meta_with_value, period_freq_str='week'):   #df_meta_with_value is only used to get metadata information
    
    statements_list = []

    statement_str_total_ooh = statement_for_total_ooh(df_for_statements, 
                                                      period_freq_str=period_freq_str)
    statements_list.append(statement_str_total_ooh)

    # preparation for the third statement

    thing_name = 'Pizza Oven'
    
    if thing_name in df_meta_with_value['thing_name'].str.lstrip().str.rstrip("0123456789 ").unique():

        building_name = df_meta_with_value['building_name'].unique()[0]
        max_period = df_meta_with_value["period"].max()
        start_time_str = max_period.start_time
        end_time_str = max_period.end_time

        statement_str_avg_action_time = statement_for_avg_action_time(db, building_name, thing_name, start_time_str, end_time_str,
                                                                      action=1, 
                                                                      period_freq_str=period_freq_str) # None will be returned if no on/off data is found
        if statement_str_avg_action_time  is not None: 

            statements_list.append(statement_str_avg_action_time)

    # Statement for biggest OOH

    # print('df_for_statements: ', df_for_statements)

    statement_str_biggest_ooh = statement_for_biggest_ooh(df_for_statements, 
                                                          period_freq_str=period_freq_str)
    statements_list.append(statement_str_biggest_ooh)
        
    return statements_list



import pandas as pd


def load_cf_from_db(engine, table_name_cache, columns):

    columns_str = ', '.join(columns)

    cache_query = f"""select {columns_str} from {table_name_cache};"""

    with engine.connect() as conn:
        df_cache = pd.read_sql_query(cache_query, con=conn)

    if df_cache.shape[0] > 0:
        df_cache.mailing_list = df_cache.mailing_list.apply(lambda x: x[1:-1].split(','))

    return df_cache




def get_sum_for_curr_and_prev(df_pivot_asset_group_by_period):

    # get the exact values of current period and period to be compared
    period_range = df_pivot_asset_group_by_period.columns
    period_current = period_range[-1]

    # for avoiding the case that we don't have data for the previous week
    if len(period_range) < 2:
        period_tobe_compared = period_range[-1]
    else:
        period_tobe_compared = period_range[-2]
        
    # get the total of current period and previous period, and the difference between the two periods
    df_pivot_asset_group_by_period_renamed = df_pivot_asset_group_by_period.loc[:,period_current].to_frame().rename(columns={period_current: "sum"})
    df_pivot_asset_group_by_period_renamed["sum_pre"] = df_pivot_asset_group_by_period.loc[:,period_tobe_compared]
    df_pivot_asset_group_by_period_renamed['sub'] = df_pivot_asset_group_by_period.loc[:,period_current] - df_pivot_asset_group_by_period.loc[:,period_tobe_compared]

    return df_pivot_asset_group_by_period_renamed




import os

# from pathlib import Path
# cache_directory = os.path.join(os.path.dirname(__file__), '_cache_/')

# credential_url = os.path.join(os.path.dirname(__file__), 'credentials.json')

credential_urls = ['/Volumes/EXT SSD/Cloudfm Backup/Documents at Cloudfm/Accounts/credentials/credentials.json',
                   '/Volumes/macOS 1/Users/Shared/Cloudfm Backup/OneDrive - Cloudfm Integrated Services Ltd - Backup/KTP Project/Cloudfm/Accounts/credentials/credentials.json']

# '/dbfs/FileStore/shared_uploads/x.yang@cloudfmgroup.com/credentials_/credentials.json'




def extract_id(x):

    # example value : '5515593330 - https://cloudfmgroup-company.monday.com/boards/5515593330'

    # if isinstance(x, float):
    #     x = int(x)

    str_x = str(x)
    if ' - ' in str_x:
        id_str = str_x.split(' - ')[0]
    else:
        id_str =  str_x
    
    if id_str == 'nan':
        return None
    else:
        return id_str.split('.')[0] # there might be the case that it's converted from float with . e.g. 8370787257.0



from datetime import date
import pandas as pd

from mdl.utils_ import get_group_with_others

def preprocessing_for_statement(df_meta_with_value, 
                                asset_group='asset_class',
                                # fixed_group_to_filter = [], # should be all in lower case charactor # applying the variable column directly
                                row_index_for_total = "Total", 
                                reading_interval_in_mins=10,
                                pct_level_tobe_others = 0):
    
    #Conversion into MWh
    w_to_kw_para = 1./1000
    min_to_hour_para = 1./60
    
    wm_to_kwh_parameter = w_to_kw_para * min_to_hour_para
    reading_to_kwh_parameter = reading_interval_in_mins * wm_to_kwh_parameter
    
    df_meta_with_value[asset_group] = df_meta_with_value[asset_group].str.strip()
    # print('df_meta_with_value: ', df_meta_with_value)

    df_meta_with_value_variable = df_meta_with_value.loc[df_meta_with_value['variable']==True]

    sr_pivot_asset_group = df_meta_with_value_variable.groupby([asset_group, 'period', 'out_of_hours'])["W"].sum() * reading_to_kwh_parameter  # Div 1000 for convertion to MWh

    # print('sr_pivot_asset_groupindex: ', sr_pivot_asset_group.index)
    sr_pivot_asset_group_unstack_ooh = sr_pivot_asset_group.unstack(["out_of_hours"])

    if True not in sr_pivot_asset_group_unstack_ooh.columns:
        return None
    else:
        df_pivot_asset_group_by_period_full = sr_pivot_asset_group_unstack_ooh[True].unstack(["period"])

        df_pivot_asset_group_by_period_no_fixed = df_pivot_asset_group_by_period_full

        # print("df_pivot_asset_group_by_period: ", df_pivot_asset_group_by_period) # debug
        
        # using the 'period' information from the dataframe
        period_range = df_pivot_asset_group_by_period_no_fixed.columns
        period_current = period_range[-1]

        # for avoiding the case that we don't have data for the previous week
        if len(period_range) < 2:
            period_tobe_compared = period_range[-1]
        else:
            period_tobe_compared = period_range[-2]

        # todo: check whether the result can be obtained by unstack directly
        # remove the assets that are generating energy
        df_pivot_asset_group_by_period = df_pivot_asset_group_by_period_no_fixed.loc[df_pivot_asset_group_by_period_no_fixed[period_current]>0].copy()
        df_pivot_asset_group_by_period_renamed = df_pivot_asset_group_by_period.loc[:,period_current].to_frame().rename(columns={period_current: "sum"})
        df_pivot_asset_group_by_period_renamed["sum_pre"] = df_pivot_asset_group_by_period.loc[:,period_tobe_compared]

        sr_total = df_pivot_asset_group_by_period_renamed.loc[df_pivot_asset_group_by_period_renamed['sum']>0].sum() # exclude the negative values

        df_total = sr_total.to_frame().transpose()
        df_total.index = [row_index_for_total]

        df_total.index.name = df_pivot_asset_group_by_period_renamed.index.name
        df_pivot_asset_group_by_period_renamed = pd.concat([df_pivot_asset_group_by_period_renamed, df_total])
        
        df_pivot_asset_group_by_period_renamed['sub'] = df_pivot_asset_group_by_period_renamed.loc[:,'sum'] - df_pivot_asset_group_by_period_renamed.loc[:,'sum_pre']

        df_pivot_asset_group_by_period_renamed["pct"] = df_pivot_asset_group_by_period_renamed["sum"]/df_pivot_asset_group_by_period_renamed.loc[row_index_for_total, "sum"]
        df_pivot_asset_group_by_period_renamed["gt_pct_others"] = df_pivot_asset_group_by_period_renamed["pct"].gt(pct_level_tobe_others)

        df_asset_group_period_sum = df_pivot_asset_group_by_period_renamed.reset_index()

        df_asset_group_period_sum["group_with_others"] = df_asset_group_period_sum.apply(get_group_with_others,asset_group=asset_group,axis=1)

        df_asset_group_period_sum_others = df_asset_group_period_sum.groupby(["group_with_others"]).sum(numeric_only=True)

        df_asset_group_period_sum_others["sum_for_sort"] = df_asset_group_period_sum_others["sum"] 

        df_asset_group_period_sum_others.loc["Others", "sum_for_sort"] = 0

        df_asset_group_period_sum_others.sort_values(["sum_for_sort"], ascending=False, inplace=True)
        df_asset_group_period_sum_others["sub_pct"] = df_asset_group_period_sum_others["sub"]/df_asset_group_period_sum_others["sum_pre"]
        
        return df_asset_group_period_sum_others

    






from .get_formatted_sum import get_formatted_sum
from .get_formatted_arrow import get_formatted_arrow
from .get_formatted_pct import get_formatted_pct
from .get_formatted_name import get_formatted_name


def generate_legend_labels(df_asset_class_monthly_sum_others, # this dataframe is required to have asset class as index and have columns "sub_pct", "sum"
                            sum_column = None,
                            pct_column = None,
                            # space_len_long = 5,
                            space_len_short = None,
                            pct_len = None,
                            kwh_len = None):
    
    if sum_column is None:
        sum_column = "sum"
    if pct_column is None:
        pct_column = "sub_pct"
    if space_len_short is None:
        space_len_short = 1

    labels = []

    # label_space_long = "," + " "*space_len_long 
    label_space_short = " "*space_len_short

    for index, item in df_asset_class_monthly_sum_others.iterrows():

        # print(f'{index = }')

        # print(f'{item = }')
        # print(f'{pct_column = }')

        label_kwh = get_formatted_sum(item[sum_column], 
                                     sum_value_len=kwh_len)
        label_arrow = get_formatted_arrow(item[pct_column])
        label_pct = get_formatted_pct(item[pct_column],
                                      pct_len=pct_len)

        label_name = get_formatted_name(index)

        label = label_kwh + label_space_short + label_arrow+ " " + label_pct + label_space_short + label_name

        # print(f'{label = }')

        labels.append(label)
        
    return labels




from .get_last_update_time import get_last_update_time
from .cache_cf_to_db import cache_cf_to_db
from .extract_contacts import extract_contacts

def sync_with_monday(engine, 
                     df_config,
                     monday_token, 
                     table_name_cache, 
                     monday_board_id,
                     columns_concerned=None, 
                     email_prefix=None):
    
    print('syncing config with monday board ...')

    # check the activity log from monday
    last_update_time = get_last_update_time(monday_token, monday_board_id)
    # print('last_update_time: ', last_update_time)

    # if activity log time > db last updates, then extract contacts from monday and update db

    flag_reload_from_monday = True

    if (df_config.shape[0] > 0) and (not df_config.last_update.isna().all()): # last_update was made null when the last_update_time was not available

        prev_update_time = df_config.last_update.max()
        print('prev_update_time: ', prev_update_time)

        if not (prev_update_time < last_update_time):

            print('[INFO]: using config from the db cache as it is up-to-date!')
            flag_reload_from_monday = False
            
    if flag_reload_from_monday:

        print('extracting contacts ... (about 3 min 30 secs)')
        df_config = extract_contacts(monday_token, monday_board_id,
                                     columns_concerned=columns_concerned, 
                                     email_prefix=email_prefix,
                                     # caching=True
                                    )
    
        # cache the result into db for the future usage and efficiency

        df_config['last_update'] = last_update_time
        # df_config = attach_id_to_buildings(engine, df_config, table_name_org, table_name_building) # df_config has to be the same name, so not renaming it here

        # print('df_config: ', df_config)
        # quit()
        cache_cf_to_db(engine, table_name_cache, df_config)

    return df_config




import matplotlib.ticker as ticker


def plot_temperature_graph(ax, sr_temp, fontsize=None, tick_range=None, label=None):

    if label is None:
        label = 'Temperature'

    if fontsize is None:
        fontsize = 13 

    # todo: month information can be removed df_occupancy

    # df_occupancy_cur.reset_index(drop=True, inplace=True)

    # the right y axis
    ax_r = ax.twinx() # instantiate a second axes that shares the same x-axis
    ax_r.yaxis.tick_left()
    ax_r.yaxis.set_label_position("left")
    ax_r.tick_params(axis='y', left=True, length=2, labelsize=fontsize)
    ax_r.set_ylabel(u'\N{DEGREE SIGN}'+'C', labelpad=5, fontsize=fontsize)
    ax_r.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1f'))
    
    if tick_range is None:
        tick_value_o_max = sr_temp.max()
        tick_value_o_min = sr_temp.min()
        tick_value_o_var = tick_value_o_max - tick_value_o_min
        tick_range = [tick_value_o_min - tick_value_o_var*0.5, tick_value_o_max + tick_value_o_var*0.5]

    ax_r.set_ylim(tick_range)
    ax_r.plot(sr_temp, color= 'k', lw=0.1, ls='dashed', marker=".", ms=6, mec="k", label=label)
    ax_r.legend(loc='upper right', bbox_to_anchor=(0.92, 0.99), fontsize=fontsize, ncol=2)

    



import matplotlib.pyplot as plt

def adjust_to_inner_ring(ax, patch_for_gap, inner_radius, outer_radius, 
                         as_inner_ring=None):

    if as_inner_ring is None:
        as_inner_ring = False
    
    if as_inner_ring:

        # remove the inner line
        patch_for_gap.set(
                edgecolor=None,
                facecolor='w',
                hatch=None,
                radius=inner_radius+0.01, # added 0.01 to completely remove/cover the black line
                gid='no legend',
                )
        # p = plt.gcf()
        # p.gca().add_artist(patch_for_gap)
        ax.add_artist(patch_for_gap)

        # add the outer line
        outer_circle = plt.Circle((0, 0), outer_radius, 
                                    fill=False, 
                                    linewidth=1, 
                                    gid='no legend',
                                    edgecolor='k') # edgecolor=other_colours[0], 
        # p = plt.gcf()
        # p.gca().add_artist(outer_circle)
        ax.add_artist(outer_circle)





def get_change_pct_and_sort_and_trim_cols(df_pivot_asset_group_by_period_renamed_rdx_to_keep,
                                            sum_column = None, 
                                            pct_column = None,
                                            gene_column = None,
                                            asset_group = None):
    
    if sum_column is None:
        sum_column = "sum_abs"

    if pct_column is None:
        pct_column = 'sub_pct'

    if gene_column is None:
        gene_column = 'generator'

    if asset_group is None:
        asset_group = 'thing_category'

    # calculate the percentage of changes
    df_pivot_asset_group_by_period_renamed_rdx_to_keep[pct_column] = df_pivot_asset_group_by_period_renamed_rdx_to_keep['sub'] / df_pivot_asset_group_by_period_renamed_rdx_to_keep['sum']

    # sort the list
    df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort = df_pivot_asset_group_by_period_renamed_rdx_to_keep.sort_values([gene_column, 'rank_in_group'])

    # the plotting function requires asset_group to be the index
    df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort_idx = df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort.set_index([asset_group])

    columns_for_export = [sum_column, pct_column, gene_column]
    df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort_idx_trim = df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort_idx[columns_for_export]

    return df_pivot_asset_group_by_period_renamed_rdx_to_keep_sort_idx_trim



#The below is the table definition
name = 'analysis.mailing_list'

columns_def_dict = {'contact_id': 'TEXT NOT NULL', 
                    'name': 'TEXT',
                    'email': 'TEXT',
                    'board_group': 'text NOT NULL', 
                    'ids_energy_report': 'text NOT NULL'}

property_dict = {"PRIMARY KEY": "(contact_id)"}

columns_dict = {**columns_def_dict, **property_dict}

hypertable = False 




def adjust_values_for_better_visibility(sr_orig,
                                        thres_invisibity=0.01,
                                        adjustment_pct=0.01):

    # ensure that the pct is positive 
    sr_orig_abs = sr_orig.abs()
    sr_plot_pct = sr_orig_abs/sr_orig_abs.sum()

    # print(f'{sr_plot_pct = }')

    # label the non-visible
    non_visible_index = sr_plot_pct.loc[sr_plot_pct<thres_invisibity].index
    all_visible_index = sr_plot_pct.index.difference(non_visible_index)

    # slightly adjust the values to make the non-visible ones more visible
    previous_non_visiable_pct = sr_plot_pct.loc[non_visible_index].sum()
    sr_plot_pct.loc[non_visible_index] = sr_plot_pct.loc[non_visible_index] + adjustment_pct
    improved_non_visiable_pct = sr_plot_pct.loc[non_visible_index].sum()

    sr_plot_pct.loc[all_visible_index] = sr_plot_pct.loc[all_visible_index] * (1-improved_non_visiable_pct)/(1-previous_non_visiable_pct)

    # get the adjusted sum values for plots
    sr_plot_sum = sr_plot_pct * sr_orig.sum()

    return sr_plot_sum




def preprocessing_for_phases(df_meta_with_value, 
                    reading_interval_in_mins=None):

    if reading_interval_in_mins is None:
        reading_interval_in_mins=10

    # Conversion into MWh
    w_to_kw_para = 1./1000
    min_to_hour_para = 1./60

    wm_to_kwh_parameter = w_to_kw_para * min_to_hour_para
    reading_to_kwh_parameter = reading_interval_in_mins * wm_to_kwh_parameter

    # consider only the consumed energy
    df_meta_with_value_positive = df_meta_with_value.loc[df_meta_with_value["W"] > 0]

    # group and pivot operation
    sr_group_phase = df_meta_with_value_positive.groupby(["phase"])["W"].sum() * reading_to_kwh_parameter

    # extract phase values to list and report possible anomalies
    existing_phases = sr_group_phase.index.to_list()
    phase_values_list = []

    for phase in ['L1', 'L2', 'L3']:
        if phase in existing_phases:
            phase_values_list.append(sr_group_phase[phase])
        else:
            print(f'[WARN]: missing phase "{phase}"!')
            phase_values_list.append(0)

    for phase in existing_phases:
        if phase not in ['L1', 'L2', 'L3']:
            print(f'[WARN]: unexpected phase "{phase}"!')

    return phase_values_list


import pandas as pd 

def launchInteractiveProcess(self, board_id:int) -> pd.DataFrame:
    data = self._fetchPrismConf(board_id)
    df = pd.DataFrame.from_records(data)
    # print(df.head(10))
    return df




def extract_id(x):

    # example value : '5515593330 - https://cloudfmgroup-company.monday.com/boards/5515593330'
    
    str_x = str(x)
    if ' - ' in str_x:
        return str_x.split(' - ')[0]
    else:
        return x



__version__ = '0.0.3' # can set multiple paths for local and databricks

# __version__ = '0.0.2' # being able to get conn from local, airflow, and databricks

# __version__ = '0.0.1' # require two functions for getting the conn

# __version__ = '0.0.0' # the basic way for getting conn only from airflow



import matplotlib.pyplot as plt

from .plot_ring import plot_ring
from .plot_centre_pct import plot_centre_pct
from .show_legend import show_legend
from .plot_phase_balance import plot_phase_balance

# cd = {"tomato": '#FF836A',
#       "aquablue": '#6DC2B3',
#       "peach": '#FED6D2',
#       "darkgrey": '#9F9D9C',
#       "potato": '#FEF8C8',
#       "cyan": '#B6E4E1',
#       "dimgray":'#696969',
#       "Seafoam":'#87e0cf',
#       "gainsboro":'#DCDCDC',
#       "grey": '#808080',
#       "brown": '#964B00',
#       "black": '#000000'}

def plot_adaptive_ring_with_phase(df_pre_pie,
                                    phase_values_list,
                                    colors=None,
                                    hatches=None,
                                    sum_column = None, 
                                    pct_column = None,
                                    gene_column = None,
                                    fontsize=None,
                                    ):
    
    if colors is None:
        from .colors import colors

    if hatches is None:
        from .hatches import hatches
    
    if sum_column is None:
        sum_column = "sum_abs"

    if gene_column is None:
        gene_column = 'generator'

    fig, ax = plt.subplots(1, 1, figsize=(9, 3.5))

    # gs = fig.add_gridspec(2, hspace=0)
    # axs = gs.subplots(sharex=True, sharey=True)

    if df_pre_pie[gene_column].nunique() == 2:

        sr_outer = df_pre_pie.loc[df_pre_pie[gene_column]==False, sum_column]
        sr_inner = df_pre_pie.loc[df_pre_pie[gene_column]==True, sum_column]
        gap_value = sr_inner.sum() - sr_outer.sum()
        outer_ring_inner_radius = 0.55
        asset_total = sr_inner.sum() + sr_outer.sum()
    else:
        sr_outer = df_pre_pie[sum_column]
        gap_value = None
        outer_ring_inner_radius = 0.50
        asset_total = sr_outer.sum()
    
    print('consumption_mwh_subchannel_sum: ', asset_total/1000) # debug

    # plot the outer ring
    plot_ring(sr_outer, ax,
                    colors=colors, 
                    fontsize=fontsize, 
                    gap_value=gap_value, 
                    outer_radius=1, 
                    inner_radius=outer_ring_inner_radius,
                    ann_pct=True, 
                    as_inner_ring=False)

    # plot the inner ring and pct text in the middle conditionally
    if df_pre_pie[gene_column].nunique() == 2:
        plot_ring(sr_inner, ax, 
                        hatch=hatches, 
                        fontsize=fontsize, 
                        gap_value=-gap_value, 
                        outer_radius=0.5, 
                        inner_radius=0.3,
                        ann_pct=False, 
                        as_inner_ring=True)

        assets_source_pct = sr_inner.sum()/sr_outer.sum()
        plot_centre_pct(ax, assets_source_pct, 
                        fontsize=fontsize)

    show_legend(ax, df_pre_pie, 
                fontsize=fontsize,
                sum_column=sum_column, 
                pct_column=pct_column)

    # ruler settings
    ru_x = 0.01  # the x-axis of phase balance ruler
    ru_y = 0  # the x-axis of phase balance ruler
    ru_w = 2  # the width of phase balance ruler
    ru_h = 0.06 # the height of phase balance ruler

    plot_phase_balance(ax, phase_values_list,
                        fontsize=fontsize,
                        ru_x=ru_x,
                        ru_y=ru_y,
                        ru_h=ru_h,
                        ru_w=ru_w)

    fig.tight_layout(pad=0, rect=[-0.2, -0.0, 0.608, 0.99])

    ax.set_ylim(-1.75, 1.6)





def exchange_group_name_column(df_meta_with_value, group_name_column_exchange, asset_group_column):

    # for the special groups where we need to use a different column_name as group_name
    for special_group_name in group_name_column_exchange:
        special_group_index = (df_meta_with_value[asset_group_column] == special_group_name)
        df_meta_with_value.loc[special_group_index, asset_group_column] = df_meta_with_value.loc[special_group_index, group_name_column_exchange[special_group_name]]

    return df_meta_with_value



import matplotlib.pyplot as plt

from .setup_x_tick_lim import setup_x_tick_lim
from .setup_y_tick_lim_label_zero_line import setup_y_tick_lim_label_zero_line
from .plot_temperature_graph import plot_temperature_graph
from .plot_the_legend import plot_the_legend

from mdl.utils_ import get_pos_bot_range_unit_format
from mdl.utils_ import plot_stacked_barchart


def plot_shifted_daily_barchart_with_temp(df_pivot_working_hours_sorted, 
                                          sr_temp=None,
                                          tick_range_temp=None, 
                                          fs=None,
                                          fontsize=None):

    if fs is None:
        fs = (8, 3.5)
    
    fig, ax = plt.subplots(1, 1, figsize=fs)

    plt.style.use('seaborn-v0_8-white')# set ggplot style

    # df_pivot_working_hours_sorted_reset_pos, \
    # sr_pivot_working_hours_sorted_reset_bot, \
    #     ytick_range = setup_y_label_and_get_pos_bot_range(ax, df_pivot_working_hours_sorted, 
    #                                                     fontsize=fontsize)
    
    df_pivot_working_hours_sorted_reset_pos, \
    sr_pivot_working_hours_sorted_reset_bot, \
        ytick_range, \
             display_unit, \
                 formatter = get_pos_bot_range_unit_format(df_pivot_working_hours_sorted, 
                                                        pad_pct_above_max=0.3)

    # plot the positive part with the bottom line shifted
    plot_stacked_barchart(ax, df_pivot_working_hours_sorted_reset_pos, 
                        bottom=sr_pivot_working_hours_sorted_reset_bot,
                        tick_range=ytick_range)

    # setup x, y tick, limit, label, and zero lines
    setup_x_tick_lim(ax, df_pivot_working_hours_sorted, 
                            fontsize=fontsize)

    # set the layout and adjust if the temperature data exists
    legend_bbox_to_anchor = None # default to: (0.5, 0.99)
    tight_layout_rect = (0, 0, 0.93, 1) # intentionally add some padding on the right hand side
    subplots_adjust_left = None
    subplots_adjust_right = 0.785 # 0.92 # 

    if sr_temp is not None:
        if sr_temp.shape[0] > 0:
            plot_temperature_graph(ax, sr_temp, 
                                fontsize=fontsize, 
                                tick_range=tick_range_temp)
            # tight_layout_rect=(0, 0, 0.94, 1)
            legend_bbox_to_anchor=(0.37, 0.99)
            # subplots_adjust_left = 0.08

    setup_y_tick_lim_label_zero_line(ax, ytick_range, display_unit, formatter,
                            fontsize=fontsize)

    # plot the legend
    plot_the_legend(ax, fontsize=fontsize, 
                    bbox_to_anchor=legend_bbox_to_anchor)

    fig.tight_layout(rect=tight_layout_rect)
    fig.subplots_adjust(left=subplots_adjust_left,
                        right=subplots_adjust_right)




def get_mailing_list(sr_contact_row):

    # df_contact_site = df_contact.loc[(df_contact.name==(site_name).replace("''", "'")) & (df_contact.org==org_name)]
    
    monday_columns = sr_contact_row.index
    email_columns = [column for column in monday_columns if 'email' in column]
    
    # df_emails_site = df_contact_site[email_columns]

    ls_emails_site = sr_contact_row[email_columns].tolist()

    # print('ls_emails_site: ', ls_emails_site)
    
    ls_emails_site_nona = [i for i in ls_emails_site if i != None]

    # ls_emails_site_nona
    # print('ls_emails_site_nona: ', ls_emails_site_nona)

    ls_emails_site_nona_with_at = [email_str for email_str in ls_emails_site_nona if '@' in email_str]

    # print('ls_emails_site_nona: ', ls_emails_site_nona)

    return ls_emails_site_nona_with_at



from cache_return import cache_return

@cache_return
def select_sites_and_emails(df_config, mode, reviewer_emails):

    if mode == 'publish':
        df_config_select = df_config.loc[df_config.publish]
        df_config_select['send_to'] = df_config_select.mailing_list
    elif mode == 'review':
        df_config_select = df_config.loc[df_config.publish]
        df_config_select['send_to'] = [reviewer_emails for _ in df_config_select.index] # assigning the same emails to every site
    elif mode == 'testing':
        df_config_select = df_config.loc[df_config.testing]
        df_config_select['send_to'] = [reviewer_emails for _ in df_config_select.index]
    else:
        raise Exception('Please provide the right mode [publish, review, testing].')
    
    return df_config_select



from typing import List
import requests

from .MondayProcessorExceptions_ import (
    MondayProcessorBoardNotFound,
    MondayProcessorNoColumInBoard,
    MondayProcessorNoItemInGroup
)


def _fetchPrismConfByBoard(self, board_id:int, columns:List[str]):
        
    column_ids = []
    
    for column_name in columns:
        for column_id in self._searchColumnIdsByName(board_id, column_name=column_name):
            column_ids.append(column_id)
    
    columns_str = '["' + '","'.join(column_ids) + '"]'
    columns_index = ["ct", "phase", "channel", "sn", "powerSupplyPhase"]
    columnsDict = dict(zip(columns,columns_index))
    headers = {"Authorization": self._token,
               "API-Version": '2024-01'}
    data = {"query": f'''query {{
        boards (ids:{board_id}){{
            id
            name
            groups(){{
                id
                title
                items_page(limit:500) {{
                    cursor
                    items{{
                        id
                        name
                        column_values(ids:{columns_str}){{
                            column{{title}}
                            text
                        }}
                }}
                }}
                
            }}
        }}
    }}'''}
    r = requests.post('https://api.monday.com/v2', headers=headers, json=data)
    data = r.json()
    if 'data' not in data:
        print(data)
    boards = data['data']['boards']
    if(len(boards) <1):
        raise MondayProcessorBoardNotFound(f"Monday board id ({board_id}) not found when perform MondayProcessor._searchConfColumnIds")
    groups = boards[0]['groups']
    if(len(groups) < 1):
        raise MondayProcessorNoColumInBoard(f"No column in board {board_id}")
    
    ret = []
    for group in groups:
        groupId = group['id']
        items = group['items_page']['items']
        if(len(items) < 1):
            raise MondayProcessorNoItemInGroup(f"No column in board {board_id}, group {groupId}")

        for item in items:
            print('item:', item)
            tmpDict = {}
            if 'Name' in columns:
                tmpDict['Name'] = item['name']
            tmpDict['building_name'] = group["title"]
            for v in item["column_values"]:
                tmpDict[v["column"]["title"]] = v["text"]
#                 key = v["id"]
#                 ai = v["additional_info"]
#                 tmpDict[key]= None if ai== None else json.loads(ai)["label"]
#             try:
#                 ch = tmpDict["channel"]
#                 chN = int(ch)
#                 tmpDict["channel"] = chN
#             except:
#                 tmpDict["channel"] = None
            ret.append(tmpDict)
    return ret


import requests

# def _fetchBoardLogsById(self, board_id:int):
#     headers = {"Authorization": self._token}
#     data = {"query": f'''query {{
#         boards (ids:{board_id}){
#             activity_logs (limit: 10) {
#             id
#             name
#         }}
#     }}'''}
#     r = requests.post('https://api.monday.com/v2', headers=headers, data=data)
#     return r.json()

def _fetchBoardLogsById(self, board_id, limit=1):
    headers = {"Authorization": self._token,
               "API-Version": '2024-01'}
    data = {"query": f'''query {{
        boards (ids:{board_id}){{
            activity_logs (limit: {limit}) {{
            id
            event
            created_at
        }}
        }}
    }}'''}
    r = requests.post('https://api.monday.com/v2', headers=headers, json=data)

    data = r.json()
    if 'data' not in data:
        print(data) 

    return data




# import os

# from pathlib import Path
# cache_directory = os.path.join(os.path.dirname(__file__), '_cache_/')

credential_urls = ['/dbfs/FileStore/shared_uploads/x.yang@cloudfmgroup.com/credentials/credentials.json']

# print('credential_url: ', credential_url)



import matplotlib.ticker as ticker

def setup_y_tick_lim_label_zero_line(ax, ls_ylim, display_unit, formatter, 
                                     freqstr=None,
                                     fontsize=None):

    if freqstr is None:
        freqstr == 'M'

    if fontsize is None:
        fontsize = 13

    # move the y_tick to the right
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")
    ax.tick_params(axis='y', which='major', pad=9, length=5, labelsize=fontsize)
    ax.set_ylim(ls_ylim)

    # set y label
    if freqstr == 'M':
        y_label = f"Monthly Usage ({display_unit})"
    else:
        y_label = f"Weekly Usage ({display_unit})"

    ax.set_ylabel(y_label, labelpad= 18, fontsize =fontsize)

    # set yaxis formatter
    ax.yaxis.set_major_formatter(ticker.FormatStrFormatter(formatter))

    # plot the zero value horizontal line
    ax.axhline(y=0, color='k', linestyle=':', zorder=-1)



def empty_table(engine, table_name, with_log=True):
    
    with engine.connect() as conn:

        # delete data (empty) in the table
        query_empty_table = f"""DELETE FROM {table_name};"""
        conn.execute(query_empty_table) 

    if with_log:
        print(f'The table [{table_name}] is empty now.')
    return



from .boards import boards

def graphql_query(board_id, group_id, column_values_ids):

    graphql_query = f'''query {{
            {boards(board_id, group_id, column_values_ids)} 
        }}'''
    
    return graphql_query
    


import requests

def _fetchBoardById(self, board_id:int):
    headers = {"Authorization": self._token,
               "API-Version": '2024-01'}
    data = {"query": f'''query {{
        boards (ids:{board_id}){{
            id
            name
        }}
    }}'''}
    r = requests.post('https://api.monday.com/v2', headers=headers, data=data)
    return r.json()



from .group_and_pivot import group_and_pivot
from .transform_asset_stats_for_adaptive_piechart import transform_asset_stats_for_adaptive_piechart

def preprocessing_for_piechart(df_meta_with_value,
                               reading_interval_in_mins=None,
                               asset_group=None,
                               minimum_cnt_for_protection=None,
                                max_display_cnt=None,
                                sum_column=None, 
                                pct_column=None,
                                gene_column=None):

    df_pivot_asset_group_by_period = group_and_pivot(df_meta_with_value, 
                                                    reading_interval_in_mins=reading_interval_in_mins,
                                                    asset_group=asset_group)

    df_sort_idx_trim = transform_asset_stats_for_adaptive_piechart(df_pivot_asset_group_by_period,
                                                                    minimum_cnt_for_protection=minimum_cnt_for_protection,
                                                                    max_display_cnt=max_display_cnt,
                                                                    sum_column=sum_column, 
                                                                    pct_column=pct_column,
                                                                    gene_column=gene_column,
                                                                    asset_group=asset_group)
    
    return df_sort_idx_trim


import pandas as pd

def load_timezone(engine, table_name):

    with engine.connect() as conn:

        query = f""" select space_id as building_id, value as timezone  from {table_name} where key ='timezone' """ # meta.space_annotations

        df_tz = pd.read_sql_query(query, con=conn) 
        
    return df_tz



import pandas as pd
from itertools import product

def patch_missing_ooh_periods_sign(df_grouped_working_hours_period,
                                  period_max,
                                  no_of_period=None):
    
    if no_of_period is None:
        no_of_period = 6

    # handle the situation that not all groups exist
    expected_sign = [True, False]
    expected_periods = [period_max-i for i in range(no_of_period)]
    expected_working_hours = [True, False]
    expected_group_index = pd.Index(product(expected_sign, expected_periods, expected_working_hours))

    missing_group_index = expected_group_index.drop(df_grouped_working_hours_period.index)

    for indice in missing_group_index:
        df_grouped_working_hours_period.loc[indice] = 0

    return df_grouped_working_hours_period



from mdl.utils_ import (
    send_email,
)
import jinja2
import pandas as pd

def email_report(receivers, smtp, email, current_period_obj, file_path_or_obj_dict):

    for receiver in receivers:

        send_to = [receiver['email']]

        environment = jinja2.Environment()
        context = {}

        period_type = current_period_obj.freqstr[0]

        if period_type == "W":
            # current_period_str = (current_period_obj).strftime("Week %W, %Y") 
            # add 1 week to the current_period_obj so that the start week is not Week 00
            period_week_str = current_period_obj.strftime('%W')
            period_year_str = current_period_obj.strftime('%Y')
            current_period_str = f'Week {(int(period_week_str)+1):02d}, {period_year_str}'

        else:
            current_period_str = current_period_obj.strftime("Month %b %Y")
        context['current_period_str'] = current_period_str
        context['receiver_name'] = receiver['name']
        

        subject_template = environment.from_string(email.subject)
        email_subject = subject_template.render(ct=context)
        
        message_template = environment.from_string(email.message)
        email_message = message_template.render(ct=context)

        # file_path_or_obj_dict = {filename: file_path/file_obj}

        send_email(smtp.send_from, send_to, 
                    email_subject, email_message, 
                    # signature=email.signature,
                    files=file_path_or_obj_dict,
                    server=smtp.server,
                    port=smtp.port,
                    username=smtp.username,
                    password=smtp.password)


from fpdf import FPDF
from PIL import ImageColor
import os

import colorsys

from mdl.utils_ import color_dict

cd = color_dict

assets_folder = os.path.join(os.path.dirname(__file__), '_assets_/')

class PDF(FPDF):

    # def __init__(self, asset_folder):
    #     FPDF.__init__(self)
    #     self.assets_folder = assets_folder

    def header(self):
        #Logo
        self.image(assets_folder+'LetterHead - header - Mindsett_weekly.png', 0, 0, self.w)
        #Fontsize and type
        self.set_font('Arial', 'B', 15)
        self.image(assets_folder+"2024_Mindsett_Logo_White.png", 10, 8, 60)
        # self.image(assets_folder+"mindsett_logo_white_transparent.png", 15, 15, 45)
        
    def footer(self):
        #Logo
        footer_height = 32
        img_path = assets_folder+'LetterHead - Footer - Mindsett_weekly.png'
        footer_width = 210
        self.image(img_path, 0, self.h-footer_height, footer_width)
        self.set_font('Arial', 'B', 15)
        
        self.set_y(self.h-footer_height + 5)
        self.set_x(20)
        self.set_font('Arial', "I", 8)
        color_rgb = ImageColor.getcolor(cd["Mindsett Blue"], "RGB")
        color_hsv = colorsys.rgb_to_hsv(*color_rgb)
        color_rgb_changed = colorsys.hsv_to_rgb(color_hsv[0], color_hsv[1], color_hsv[2])
        self.set_text_color(*color_rgb_changed)
        #self.cell(pdf.w - 30, 10, '**BBP benchmarking (REEB) ', 0, 0, 'B')
        
    from .write_multicell_with_styles import write_multicell_with_styles


# import smtplib
# from pathlib import Path
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from email import encoders
import io

from .get_smtp_server import get_smtp_server

def send_email(send_from, send_to, subject, message, signature=None, cc=None, bcc=None, files={},
              server="localhost", port=587, username='', password='',
              use_tls=True):
    """Compose and send email with provided info and attachments.

    Args:
        send_from (str): from name
        send_to (list[str]): to name(s)
        subject (str): message title
        message (str): message body
        files (dict{str:str|obj}): list of file paths to be attached to email
        server (str): mail server host name
        port (int): port number
        username (str): server auth username
        password (str): server auth password
        use_tls (bool): use TLS mode
    """
    
    msg = MIMEMultipart()
    msg['From'] = send_from
    msg['To'] = COMMASPACE.join(send_to)
    
    if cc is not None:
        msg['Cc'] = COMMASPACE.join(cc)
        
    if bcc is not None:
        msg['Bcc'] = COMMASPACE.join(bcc)
        
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = subject

    html_msg = MIMEText(message, 'html')
    msg.attach(html_msg)

    if signature != None:
        sgn = MIMEText(signature, "html")
        msg.attach(sgn)

    for filename in files:
        part = MIMEBase('application', "octet-stream")

        file_path_obj = files[filename]

        if isinstance(file_path_obj, io.BytesIO):
            part.set_payload(file_path_obj.getvalue())
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            f'attachment; filename={filename}')
        else:
            with open(file_path_obj, 'rb') as file:
                part.set_payload(file.read())
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            f'attachment; filename={filename}')
        msg.attach(part)

    smtp = get_smtp_server(server, port)
    if use_tls:
        smtp.starttls()
    smtp.login(username, password)
    smtp.sendmail(send_from, send_to, msg.as_string())
    smtp.quit()
     
    print(f'[smtp]: email sent successfully to {str(send_to)}!')



from .extract_contacts import extract_contacts
from .cache_cf_to_db import cache_cf_to_db

def sync_with_sharepoint(engine, df_cfg_db, sharepoint, table_name_cache):

    df_cfg_shp = extract_contacts(sharepoint) # currently, we are getting the last_update time by directly querying the file, as it is relative quick
    # if needed in the future, a dedicated lighter get_last_update_time function can be designed, and the actual extraction processing can be conditional

    flag_reload_from_sharepoint = True

    if (df_cfg_db.shape[0] > 0) and (not df_cfg_db.last_update.isna().all()): # last_update was make null when the last_update_time was not available

        last_update_db  = df_cfg_db['last_update'].max()
        last_update_shp = df_cfg_shp['last_update'].max()
        # print(f'{last_update_shp=}')

        if last_update_db == last_update_shp: # here some efforts of writing to db can be saved if the cached data is already up-to-date.
            
            print('[INFO]: using config from the db cache as it is up-to-date!')
            flag_reload_from_sharepoint = False
        
    if flag_reload_from_sharepoint:

        # pass
        cache_cf_to_db(engine, table_name_cache, df_cfg_shp)

    return df_cfg_shp




def adjust_gap_for_visibility(gap_value, non_gap_value, 
                              min_vis_pct=None):
    
    if min_vis_pct is None:
        min_vis_pct = 3

    min_non_gap_pct = min_vis_pct
    max_non_gap_pct = 100 - min_non_gap_pct

    max_gap_for_visibility = (100 - min_non_gap_pct)*non_gap_value/min_non_gap_pct
    min_gap_for_visibility = (100 - max_non_gap_pct)*non_gap_value/max_non_gap_pct

    adj_gap_for_visibility = gap_value
    adj_gap_for_visibility = min(max_gap_for_visibility, adj_gap_for_visibility)
    adj_gap_for_visibility = max(min_gap_for_visibility, adj_gap_for_visibility)

    return adj_gap_for_visibility


from os import path
import pandas as pd

from .query_building_total import query_building_total
from .import_data_with_meta_postgres import import_data_with_meta_postgres

from cache_return import cache_return

@cache_return
def load_source_data(postgresdb, 
                     start_time, 
                     start_time_co2_barchart,
                     end_time, 
                     building_id,
                     exception=None,
                     chunksize=10,
                     timezone='UTC'):

    df_meta_with_value = import_data_with_meta_postgres(postgresdb, 
                                                            start_time, 
                                                            end_time, 
                                                            building_id,
                                                            exception=exception,
                                                            chunksize=chunksize,
                                                            timezone=timezone)
    df_meta_with_value_building = query_building_total(postgresdb, 
                                                        start_time=start_time_co2_barchart,
                                                        end_time=end_time, 
                                                        building_id=building_id, # please note that this is a postition arg
                                                        timezone=timezone)
    
    return df_meta_with_value, df_meta_with_value_building




def get_formatted_name(name_text, 
                       max_len=None):
    
    if max_len is None:
        max_len = 15

    if len(name_text) > max_len:
        label_name = name_text[:(max_len-3)]+"..."
    else:   
        label_name = name_text

    return label_name




def get_rank_overall_protect(df_pivot_asset_group_by_period_renamed, 
                             minimum_cnt_for_protection=None):

    if minimum_cnt_for_protection is None:
        minimum_cnt_for_protection = 3

    # get the overall ranking considering the minimum entries to protect
    df_pivot_asset_group_by_period_renamed['protect_in_group'] = (df_pivot_asset_group_by_period_renamed['rank_in_group'] <= minimum_cnt_for_protection)
    df_pivot_asset_group_by_period_renamed_rdx = df_pivot_asset_group_by_period_renamed.sort_values(by=['protect_in_group', 'sum_abs'], ascending=False).reset_index()
    df_pivot_asset_group_by_period_renamed_rdx['rank_overall_protect'] = df_pivot_asset_group_by_period_renamed_rdx.index + 1

    return df_pivot_asset_group_by_period_renamed_rdx



from ..MondayProcessorException import MondayProcessorException

class MondayProcessorBoardColumnNotMatch(MondayProcessorException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from .check_table_existence import check_table_existence
from .create_new_schema import create_new_schema
from .empty_table import empty_table
from .create_table import create_table

def setup_table(engine, table_name, columns_dict, 
                hypertable=False, 
                index=None,
                constraint=None,
                redo_from_start=False, 
                with_log=True):
    """
    Function Name: setup_table
    Parameters   : sql connection engine, table name, redo_from_start flag
    Description  : if redo from start set inconfiguration file - emptys table 
                   otherwise creates a new table and schema if required 
    """
    
    is_exist = check_table_existence(engine, table_name)
    
    if is_exist:
        if redo_from_start:
            empty_table(engine, table_name, 
                        with_log=with_log)
    else:
        table_schema_name_list = table_name.split(".")

        if len(table_schema_name_list) > 1:

            # TODO: add check_schema_existence
            table_schema = table_schema_name_list[0]
            create_new_schema(engine, table_schema)
        
        create_table(engine, table_name, columns_dict, 
                     constraint=constraint,
                     index=index,
                     hypertable=hypertable)
        
    return



import numpy as np

def plot_stacked_barchart(ax, df_pivot_working_hours, bottom=None,
                          tick_range=None,
                          top_hours=None):

    if tick_range is None:
        tick_value_e_max = max(0, df_pivot_working_hours.sum(axis=1).max())
        tick_value_e_min = min(0, df_pivot_working_hours.sum(axis=1).min())
        tick_range = [tick_value_e_min, tick_value_e_max]

    if bottom is None:
        bottom = 0
    if top_hours is None:
        top_hours = True

    tick_value_e_var = tick_range[1] - tick_range[0]
    white_padding_below_bar = abs(tick_value_e_var/100)

    bar_padding_adjustment = 0.01 # used when the paddings on the left/right are different
    
    bot_hours = not top_hours

    bar_edgecolour = ['k','w']
    bar_fillcolour = ['k','w']
    bar_color = '#6DC2B3'

    hours_labels = {True: "Out Of Hours", False: "In Hours"}
    hours_colors = {True: "w", False: bar_color}

    # white edge/pad of bar
    ax.bar(df_pivot_working_hours.index, df_pivot_working_hours[top_hours].fillna(0)+df_pivot_working_hours[bot_hours].fillna(0),
           bottom=bottom,
           width=0.7, 
           lw=1.3, 
           edgecolor=bar_edgecolour[0], 
           color=bar_fillcolour[1], 
           label=hours_labels[top_hours])

    # bottom bar inner part
    inner_bar_bot_pad = tick_value_e_var*0.01
    df_white_padding_below_bar_conditional = np.sign(df_pivot_working_hours[bot_hours])*(white_padding_below_bar+inner_bar_bot_pad)
    ax.bar(df_pivot_working_hours.index+bar_padding_adjustment, df_pivot_working_hours[bot_hours].fillna(0) - df_white_padding_below_bar_conditional,
           bottom=bottom+inner_bar_bot_pad,
           width=0.4, 
           lw=0, 
           color= hours_colors[bot_hours], 
           edgecolor=bar_edgecolour[1], 
           label=hours_labels[bot_hours])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#author: Xiao Wang 
# x.wang@cloudfmgroup.com
# Created: 05/04/2023

from abc import ABC

class DataProcessor(ABC):

    from .launchInteractiveProcess import launchInteractiveProcess

    def __init__(self) -> None:
        super().__init__()




def get_subgroup_sum_and_rank(df_pivot_asset_group_by_period_renamed,
                              sum_column = None, 
                              gene_column = None):
    
    if sum_column is None:
        sum_column = "sum_abs"

    if gene_column is None:
        gene_column = 'generator'

    # label the entries based on whether generator or not and add sub-group total
    df_pivot_asset_group_by_period_renamed[gene_column] = df_pivot_asset_group_by_period_renamed['sum'] < 0
    # df_pivot_asset_group_by_period_tot = df_pivot_asset_group_by_period_renamed.groupby([gene_column])['sum'].sum().rename('generator_tot')
    # df_pivot_asset_group_by_period_renamed = df_pivot_asset_group_by_period_renamed.merge(df_pivot_asset_group_by_period_tot, left_on=[gene_column], right_index=True)

    # get the rank in sub-group
    df_pivot_asset_group_by_period_renamed[sum_column] = df_pivot_asset_group_by_period_renamed['sum'].abs()
    df_pivot_asset_group_by_period_renamed['rank_in_group'] = df_pivot_asset_group_by_period_renamed.groupby(gene_column)[sum_column].rank(ascending=False, method="dense")

    return df_pivot_asset_group_by_period_renamed






def groupby_date_and_ooh_current_period(df_meta_with_value):

    # select data only for the current period

    period_current = df_meta_with_value["period"].max()
    df_meta_with_value_current = df_meta_with_value.loc[df_meta_with_value["period"]==period_current]

    df_meta_with_value_current['neg_sign'] = df_meta_with_value_current['W'].lt(0)

    # Conversion into MWh
    w_to_mw_para = 1./1000/1000
    min_to_hour_para = 1./60
    reading_interval_in_mins=10

    wm_to_mwh_parameter = w_to_mw_para * min_to_hour_para
    reading_to_mwh_parameter = reading_interval_in_mins * wm_to_mwh_parameter

    # Conversion into MWh
    # reading_to_mwh_parameter = 1./1000

    # todo: Better to change to df_meta_value_building
    df_grouped_working_hours_current_period_with_sign = df_meta_with_value_current.groupby(['neg_sign', "date", 'out_of_hours'])["W"].sum() * reading_to_mwh_parameter  # Div 1000 for convertion to MWh

    # df_grouped_working_hours_current_period = df_grouped_working_hours_current_period_with_sign.droplevel('neg_sign')
    # # select only the maximum period
    # df_grouped_working_hours_multiple_period_unstack = df_grouped_working_hours_multiple_period.unstack(["period"])
    # period_current = df_grouped_working_hours_multiple_period_unstack.columns[-1]
    # df_grouped_working_hours = df_grouped_working_hours_multiple_period_unstack.loc[:, period_current]

    # todo: fill missing group index
    df_pivot_working_hours_with_sign_sorted = df_grouped_working_hours_current_period_with_sign.unstack(['neg_sign', 'out_of_hours']).sort_index(axis=1, 
                                                                                                                 level=1, 
                                                                                                                 ascending=False)
    
    return df_pivot_working_hours_with_sign_sorted, period_current




def get_meta_statement(building_id, 
                    #    organisation=None, 
                       exception=None,
                      ):

    statement_list = [f"""root_id='{building_id}'""", "thing_id is not null", "thing_name is not null"] # please note that the mt. lt. shortname for tables are still coupled with function load_metadata. Update - 20240814, resolved.

    # if organisation != None:
    #     statement_new  = f"""btrim(org) = '{organisation}'"""
    #     statement_list.append(statement_new)

    if exception != None:
        for key in exception:
            is_list = isinstance(exception[key], list) 
            
            if is_list:
                exc_list = exception[key]
            else:
                exc_list = [exception[key]]
            
            exc_list_with_quote = [f"'{item}'" for item in exc_list]

            exc_concat_str = ', '.join(exc_list_with_quote)

            statement_new  = f"""{key} not in ({exc_concat_str})"""
            statement_list.append(statement_new)

    statement_full = " and ".join(statement_list)

    return statement_full


from pathlib import Path
import pickle

from .preprocessing_for_statement import preprocessing_for_statement
from .insight_statements import insight_statements


def generate_insight_statements(db, df_meta_with_value, 
                                asset_group = "asset_class",
                                # fixed_group_to_filter = [],
                                directory_to_savefile=None): # todo: update the default value for directory

    df_for_statements = preprocessing_for_statement(df_meta_with_value, asset_group=asset_group)

    # print('df_for_statements: ', df_for_statements) # debug

    if df_for_statements is not None:
        # get the readable name of period frequency
        current_period_obj = df_meta_with_value.period.max()
        period_type = current_period_obj.freqstr[0]
        period_freq_str_dict = {'W': 'week', 'M': 'month'}
        period_freq_str = period_freq_str_dict[period_type]

        statements_list = insight_statements(db, df_for_statements, df_meta_with_value, period_freq_str=period_freq_str)
    else:
        statements_list = ['Auto observations for 24/7 sites are not yet available!']

    # Specify the directory to save figures, if it does not exist, create it
    # Path(directory_to_savefile).mkdir(parents=True, exist_ok=True)

    # with open(directory_to_savefile+'statements.pkl', 'wb') as f:
    #     pickle.dump(statements_list, f)

    pkl_name = 'statements.pkl'
    if directory_to_savefile == None:
        # obj_object = io.BytesIO()
        return {pkl_name: statements_list}
    else:
        pkl_path = directory_to_savefile + pkl_name
        with open(pkl_path, 'wb') as f:
            pickle.dump(statements_list, f)
        return {pkl_name: pkl_path}



columns_concerned = ['last_update', 'building_id', 'item_name', 'group_name', 'board_id',
       'hvac_board_id', 'publish', 'testing', 'industry', 'manager_name',
       'currency', 'occupancy_available', 'asset_group', 'fillna_value',
       'insight_statements', 'group_name_column_exchange',
       'group_name_modification', 'pct_level_tobe_others', 'floor_sqm', 'pct_hide',
       'conv_mwh_price', 'period_freq', 'mailing_list']


def append_df_to_table(engine, table_name, df):
    
    table_schema_name_list = table_name.split(".")

    if len(table_schema_name_list) > 1:
        table_schema = table_schema_name_list[0]
        table_non_schema_name = table_schema_name_list[1]
    else:
        table_schema = "public"
        table_non_schema_name = table_name
        
    total_records = len(df)
    
  
    with engine.connect() as conn:
  
        df.to_sql(
           name=table_non_schema_name,
           con=conn,
           if_exists="append",
           schema=table_schema,
           chunksize=5000,
           index=False
           )

    print(f'The config cache in database table [{table_name}] is updated with [{total_records}] records.')
    # print(f'[{total_records}] records have been written to Table [{table_name}] successfully.')
   
    return 



from matplotlib.offsetbox import OffsetImage, AnnotationBbox

import matplotlib.image as mpimg
import matplotlib.patches as mpatches
import matplotlib.text as mpltext

from os import path

cd = {"tomato": '#FF836A',"aquablue": '#6DC2B3',"peach": '#FED6D2',"darkgrey": '#9F9D9C',"potato": '#FEF8C8',
      "cyan": '#B6E4E1',"dimgray":'#696969',"Seafoam":'#87e0cf',"gainsboro":'#DCDCDC'}

def energy_meter_design(ax, consumption_mwh_cur, consumption_mwh_pre, image_folder,
                        conv_mwh_co2 = 0.233,
                        conv_mwh_price = 190,
                        fx = 0.2, # fancy box x-axis value
                        fy = 0.3, # fancy box y-axis value
                        fs  = 0.6, # fancy box scale factor
                        currency = 'GBP',
                        fontsize = 13
                        ):

    ax.set_xlim(0.2, 0.9)
    ax.set_ylim(0.2, 0.9)
    iz  = 0.06 * fs # icon zooming factor

    # add a fancy box as the energy meter outline

    fancybox_width  = fs
    fancybox_height = fs
    fancybox = mpatches.FancyBboxPatch(
                [fx, fy], fancybox_width, fancybox_height, fc="w", ec='k', lw=1.5, ls= 'dotted',
                boxstyle=mpatches.BoxStyle("Round", rounding_size=0.07, pad=0.02))
    ax.add_artist(fancybox)

    notes_fontsize = fontsize - 2
    # text sub-title 
    text_title = mpltext.Text(x=fx+0.45*fs, y=fy+0.7*fs, text=f'Equivalent$^*$', 
                              va=u'baseline', color='grey', ha=u'right', fontweight='light',fontstyle='italic', fontsize=notes_fontsize)
    ax.add_artist(text_title)

    # text footnote

    ## currency_symbol selection
    currency_symbol_dict = {
        'GBP': '£',
        'EUR': '€',
        'USD': '\$',
        'SGD': 'S\$'
    }
    try: 
        currency_symbol = currency_symbol_dict[currency]
    except:
        raise Exception('There is no symbol provided for this currency!')

    ## symbol selection
    text_footnote_with_co2 = f'* {conv_mwh_co2:1.3f}$t$ CO2e $\Leftrightarrow$ 1MWh $\Leftrightarrow$ {currency_symbol}{conv_mwh_price:1.1f}' # not used for now - 2024 Sep 23th
    text_footnote_without_co2 = f'* 1MWh $\Leftrightarrow$ {currency_symbol}{conv_mwh_price:1.1f}'
    text_footnote = mpltext.Text(x=fx+0.5*fs, y=fy+0.04*fs, text=text_footnote_without_co2, 
                                 va=u'baseline', color='k', ha=u'center', fontsize=notes_fontsize)
    ax.add_artist(text_footnote)

    ## text readings settings
    re_w = 0.80   # reading width
    re_x = 0.127  # reading x-axis value
    re_y = 0.80   # reading y-axis value
    re_c = "k"    # reading colour
    r_fs = fontsize + 3   # reading fontsize # please note that the other sizes except default 13 haven't been tested yet
    mr_fs = fontsize + 5    # main reading fontsize # please note that the other sizes except default 13 haven't been tested yet

    ## reading - energy 
    reading = consumption_mwh_cur

    reading_abs = abs(reading) # the default unit is MWh

    # adjust the decimal point and the unit based on the actual value
    if reading_abs > 10:
        reading_value = f"{reading:.0f} MWh"
    elif reading_abs > 1:
        reading_value = f"{reading:.1f} MWh"
    elif reading_abs*1000 > 10:
        reading_value = f"{reading*1000:.0f} KWh"
    elif reading_abs*1000 > 1:
        reading_value = f"{reading*1000:.1f} KWh"
    elif reading_abs*1000_000 > 10:
        reading_value = f"{reading*1000_000:.0f} Wh"
    else:
        reading_value = f"{reading*1000_000:.1f} Wh"
    # reading - energy - text - monthly total
    rt_x = re_x + 0.15 + 0.03
    rt_y = re_y
    readingtext = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{reading_value}', 
                               va=u'bottom', color=re_c, ha=u'center', fontweight="bold", fontsize=mr_fs)

    # reading - energy - text - monthly change
    rt_x = re_x + 0.65
    rt_y = re_y + 0.02


    change_value = (consumption_mwh_cur - consumption_mwh_pre) / abs(consumption_mwh_pre)
    change_value_int = int(change_value*100)

    # print(f'{consumption_mwh_cur = }')
    # print(f'{consumption_mwh_pre = }')
    # print(f'{change_value = }')


    if change_value_int > 0: 
        change_arrow_str = r'${\blacktriangle}$'
    elif change_value_int < 0:
        change_arrow_str = r'$\:\!\triangledown\:\!$'
    else :
        change_arrow_str = r'--'


    change_in_percentage_value = int(abs(change_value)*100)

    if change_in_percentage_value > 999:
        change_in_percentage_value_str = r"$\,$" + ">999" + "%"
    elif change_in_percentage_value > 99: 
        change_in_percentage_value_str = r"$\,$" + str(change_in_percentage_value) + "%"
    elif change_in_percentage_value > 9: 
        change_in_percentage_value_str = " " + str(change_in_percentage_value) + r"$\,$" + "%"
    else:
        change_in_percentage_value_str = " " + str(change_in_percentage_value) + r"$\;$" + "%"


    change_in_percentage = change_arrow_str + change_in_percentage_value_str
    text_percentage = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{change_in_percentage}', 
                                   va=u'baseline', color='k', ha=u'center', fontweight="bold", fontsize=mr_fs)
    ax.add_artist(text_percentage)

    # horizontal line to separate the different readings
    rl_x = re_x * 0.2       # the x-axis value of the line at the bottom of the reading 
    rl_y = re_y - 0.15        # the y-axis value of the line at the bottom of the reading
    rl_w = re_w * 1.2    # the width value of the line at the bottom of the reading
    horizontal_line = mpatches.Rectangle([fx+rl_x*fs, fy+rl_y*fs], 
                                         rl_w*fs, 0, facecolor="k", edgecolor='k', lw=1, ls= 'dotted')
    ax.add_artist(readingtext)
    ax.add_artist(horizontal_line)


    ## reading - co2
    re_y = 0.48
    reading = consumption_mwh_cur * conv_mwh_co2

    # adjust the unit based on value
    reading_value = f"{reading:1.1f} tons"
    if reading_value == '0.0 tons':
        reading_value = f"{round(reading*1000)} kg"
        
    image_name = 'co2.png'

    # reading - co2 - icon
    ri_x = re_x + 0.09  # reading icon x-axis value
    ri_y = re_y + 0.042
    icon = mpimg.imread(image_folder + image_name)
    ibox = OffsetImage(icon, zoom=2*iz)
    readingicon = AnnotationBbox(ibox, (fx+ri_x*fs, fy+ri_y*fs), frameon = False)

    # reading - co2 - text
    rt_x = re_x + 0.342
    rt_y = re_y
    readingtext = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{reading_value}', 
                               va=u'bottom', color=re_c, ha=u'left', fontweight="bold", fontsize=r_fs)

    # horizontal line to separate the different readings
    rl_x = re_x         # the x-axis value of the line at the bottom of the reading 
    rl_y = re_y - 0.07      # the y-axis value of the line at the bottom of the reading
    rl_w = re_w * 0.9   # the width value of the line at the bottom of the reading
    horizontal_line = mpatches.Rectangle([fx+rl_x*fs, fy+rl_y*fs], 
                                         rl_w*fs, 0, facecolor="k", edgecolor='k', lw=1, ls= 'dotted')

    ax.add_artist(readingicon)
    ax.add_artist(readingtext)
    ax.add_artist(horizontal_line)


    ## reading - billing
    re_y = 0.23
    reading = consumption_mwh_cur * conv_mwh_price
    reading_value = f"{reading:,.2f}".replace(".", ". ") 

    # adjust the size for visual effect
    if len(reading_value) > 8:
        reading_value = reading_value.replace(". ", ".") 
        if len(reading_value) > 8:
            reading_value = reading_value[:-2]

    # reading - billing - icon

    ## icon selection
    currency_icon_dict = { # please note that the format has to be 512px png
        'GBP': 'pound-sterling.png',
        'EUR': 'euros.png',
        'USD': 'dollar-symbol.png', # https://www.flaticon.com/free-icon/dollar-symbol_1140418?term=dollar+symbol&page=1&position=22&origin=search&related_id=1140418
        'SGD': 'singapore-dollar.png' # https://www.flaticon.com/free-icon/singapore-dollar_9921422?term=singapore+dollar&page=3&position=73&origin=search&related_id=9921422
    }
    try: 
        image_name = currency_icon_dict[currency]
    except:
        raise Exception('There is no icon provided for this currency!')

    ## icon plotting
    ri_x = re_x + 0.09  # reading icon x-axis value
    ri_y = re_y + 0.06
    icon = mpimg.imread(image_folder + image_name)
    ibox = OffsetImage(icon, zoom=1.6*iz)
    readingicon = AnnotationBbox(ibox, (fx+ri_x*fs, fy+ri_y*fs), frameon = False)

    # reading - billing - text
    rt_x = re_x + 0.342
    rt_y = re_y
    readingtext = mpltext.Text(x=fx+rt_x*fs, y=fy+rt_y*fs, text=f'{reading_value}', 
                               va=u'bottom', color=re_c, ha=u'left', fontweight="bold", fontsize=r_fs)

    # horizontal line to separate the different readings
    rl_x = re_x          # the x-axis value of the line at the bottom of the reading 
    rl_y = re_y - 0.07        # the y-axis value of the line at the bottom of the reading
    rl_w = re_w * 0.9    # the width value of the line at the bottom of the reading
    horizontal_line = mpatches.Rectangle([fx+rl_x*fs, fy+rl_y*fs], 
                                         rl_w*fs, 0, facecolor="k", edgecolor='k', lw=1, ls= 'dotted')

    ax.add_artist(readingicon)
    ax.add_artist(readingtext)
    ax.add_artist(horizontal_line)


    ## benchmarking design
    re_y = 0.65
    ri_y = re_y + 0.07*fs

    ax.axis('equal')




def group_and_pivot(df_meta_with_value, 
                    asset_group=None,
                    reading_interval_in_mins=None):

    if asset_group is None:
        asset_group = 'thing_category'

    if reading_interval_in_mins is None:
        reading_interval_in_mins=10

    # Conversion into MWh
    w_to_kw_para = 1./1000
    min_to_hour_para = 1./60

    wm_to_kwh_parameter = w_to_kw_para * min_to_hour_para
    reading_to_kwh_parameter = reading_interval_in_mins * wm_to_kwh_parameter

    # group and pivot operation
    sr_pivot_asset_class = df_meta_with_value.groupby([asset_group, "period"])["W"].sum() * reading_to_kwh_parameter
    df_pivot_asset_group_by_period = sr_pivot_asset_class.unstack(["period"])

    return df_pivot_asset_group_by_period


import pandas as pd

def load_sqm(engine, table_name):

    with engine.connect() as conn:

        query = f""" select space_id as building_id, sqm as floor_size  from {table_name};""" # meta.space_annotations

        df_tz = pd.read_sql_query(query, con=conn) 
        
    return df_tz



import sys

sys.path.append('../')

import libs
import mdl
import config as cf


param_list = sys.argv[1:]
param_dict = {}
for i in range(len(param_list)//2):
    param_dict[param_list[int(i*2)]] = param_list[int(i*2+1)]

# redo_from_start = bool(dbutils.widgets.get("redo_from_start"))

print('param_dict: ', param_dict)

mode_param = param_dict['mode']

print('mode_param: ', mode_param)

# mode = 'publish'
cache_directory = '/dbfs/FileStore/shared_uploads/x.yang@cloudfmgroup.com/cache_folder_for_jobs/'

mdl.prepare_and_email_reports(cf, 
                              mode=mode_param, 
                              sync_source=True,
                              cache_in_memory=True, 
                              cache_directory=cache_directory)



columns_concerned = ['Name', 
                    'publish', 
                    'testing', 
                    'industry', 
                    'manager_name', 
                    'currency', 
                    'occupancy_available', 
                    'asset_group', 
                    'fillna_value', 
                    'insight_statements', 
                    'group_name_column_exchange', 
                    'group_name_modification', 
                    'pct_level_tobe_others', 
                    'floor_sqm',
                    'pct_hide', 
                    'conv_mwh_price', 
                    'period_freq',
                    'building_id',
                    'board_id',
                    'hvac_board_id'
                    ]




def get_excel_file_from_link(ctx, sharepoint_link):

    # get the folder link from the file link
    sharepoint_folder_link = sharepoint_link.rsplit('/', maxsplit=1)[0]
    
    # get sharepoint folder
    libraryRoot = ctx.web.get_folder_by_server_relative_path(sharepoint_folder_link)
    ctx.load(libraryRoot)
    ctx.execute_query()

    # get files in folder
    files_in_folder = libraryRoot.files
    ctx.load(files_in_folder)
    ctx.execute_query()

    # debug
    # print(f'{files_in_folder = }')

    # with the condition that there is only one file
    # excel_file = files_in_folder[0]

    for file in files_in_folder:
        files_url = file.properties['ServerRelativeUrl']
        if files_url == sharepoint_link:
            return file

    raise Exception('[ERROR]: Cannot find the specified file!')
    # return excel_file


import matplotlib.pyplot as plt
plt.rcParams['figure.max_open_warning'] = 300

from .generate_insight_statements import generate_insight_statements
from .generate_piechart_with_phase import generate_piechart_with_phase
from .generate_energy_meter_with_benchmarking import generate_energy_meter_with_benchmarking
from .generate_daily_barchart_with_temp import generate_daily_barchart_with_temp
from .generate_period_barchart import generate_period_barchart

def generate_components(postgresdb, df_meta_with_value, df_meta_with_value_building, site_obj, 
                        fontsize=None, 
                        directory_to_savefile=None):

    insight_statements = generate_insight_statements(postgresdb, df_meta_with_value, 
                                asset_group=site_obj.asset_group, 
                                # fixed_group_to_filter=site_obj.fixed_group_to_filter,
                                directory_to_savefile=directory_to_savefile)
    
    piechart_with_phase = generate_piechart_with_phase(df_meta_with_value, 
                                fontsize=fontsize,
                                directory_to_savefig=directory_to_savefile)
    
    energy_meter_with_benchmarking = generate_energy_meter_with_benchmarking(df_meta_with_value_building, site_obj.floor_size, 
                                            size_in_sqm=site_obj.size_in_sqm, 
                                            conv_mwh_price=site_obj.conv_mwh_price,
                                            industry=site_obj.industry, 
                                            currency=site_obj.currency,
                                            fontsize=fontsize,
                                            directory_to_savefig=directory_to_savefile)

    daily_barchart_with_temp = generate_daily_barchart_with_temp(df_meta_with_value,
                                                                fontsize=fontsize,
                                                                directory_to_savefig=directory_to_savefile)

    period_barchart = generate_period_barchart(df_meta_with_value_building,
                                         fontsize=fontsize,
                                         directory_to_savefig=directory_to_savefile)
    
    # if directory_to_savefile == None:
    components = {
        **insight_statements,
        **piechart_with_phase,
        **energy_meter_with_benchmarking,
        **daily_barchart_with_temp,
        **period_barchart
        }
    return components


from .column import column

def column_values(ids):
    column_values = f"""
                column_values (ids:{ids}) {{
                            {column}
                            text
                        }}
                        """
    return column_values



import pandas as pd

def load_iot_data_partition(engine, time_range, iot_table_name, id_list_part):

    time_condition = f""" ((time >= '{time_range[0]}') and (time < '{time_range[1]}')) """

    id_list_str = str(tuple(id_list_part)).replace('),)', '))')

    data_query = f"""select time, id, kwh*6000 as W from {iot_table_name} 
                where id in {id_list_str}
                    and {time_condition};"""

    # print('data_query: ', data_query)

    with engine.connect() as conn:
        df_iot_part_iter = pd.read_sql_query(data_query, con=conn, chunksize=5000)

    df_iot_part = pd.DataFrame([])

    for df_iot_part_i in df_iot_part_iter:
        df_iot_part = pd.concat([df_iot_part, df_iot_part_i])

    return df_iot_part




def get_phase_markers(phase_values_list, 
                      var_allowance_pct=None):

    balanced_avg_value = sum(phase_values_list)/3

    if var_allowance_pct is None:
        var_allowance_pct = 0.04

    var_allowance_value = balanced_avg_value * var_allowance_pct

    markers_list = []
    for phase_value in phase_values_list:

        if phase_value > (balanced_avg_value + var_allowance_value):
            marker_text = '>'
        elif phase_value > (balanced_avg_value - var_allowance_value):
            marker_text = '='
        elif phase_value > 0:
            marker_text = '<'
        else:
            marker_text = '?'

        markers_list.append(marker_text)
    
    return markers_list
        



from .append_df_to_table import append_df_to_table
from .empty_table import empty_table

def cache_cf_to_db(engine, table_name_cache, df_config):

    empty_table(engine, table_name_cache, with_log=False)
    append_df_to_table(engine, table_name_cache, df_config)

    return





from .groups import groups

def boards(id, group_id, column_values_ids):
    boards = f'''
    boards (ids:{id}){{
                id
                name
                {groups(group_id, column_values_ids)}
            }}
    '''
    return boards



# modify the names in asset_group based on the provided dictionary

def modify_group_names(df_meta_with_value, group_name_column="asset_type", modification_dict={}):
    
    # the letter cases and spaces prefix/sufix will not affect the result
    df_meta_with_value[group_name_column] = df_meta_with_value[group_name_column].str.strip()

    modification_dict_with_lowercase_key = {}
    for group in modification_dict:
        modification_dict_with_lowercase_key[group.lower().strip()] = modification_dict[group]

    df_meta_with_value[group_name_column] = df_meta_with_value[group_name_column] \
                                                                            .str.lower() \
                                                                            .map(modification_dict_with_lowercase_key) \
                                                                            .fillna(df_meta_with_value[group_name_column])
    return df_meta_with_value


from .patch_missing_dates import patch_missing_dates
from .patch_missing_columns_for_ooh_and_sign import patch_missing_columns_for_ooh_and_sign
from .groupby_date_and_ooh_current_period import groupby_date_and_ooh_current_period

# from .groupby_date_and_ooh_current_period import groupby_date_and_ooh_current_period

from cache_return import cache_return


@cache_return
def preprocessing_for_barchart(df_meta_with_value_bld):
    
    # main operation
    df_pivot_working_hours_with_sign_sorted, period_current = groupby_date_and_ooh_current_period(df_meta_with_value_bld)

    # patch the missing ooh columns and period index
    df_pivot_working_hours_with_sign_sorted = patch_missing_columns_for_ooh_and_sign(df_pivot_working_hours_with_sign_sorted)
    df_pivot_working_hours_with_sign_sorted = patch_missing_dates(df_pivot_working_hours_with_sign_sorted, period_current)

    return df_pivot_working_hours_with_sign_sorted, period_current
    # pass



import matplotlib.pyplot as plt



def plot_centre_pct(ax, assets_source_pct, 
                    fontsize=None):

    if fontsize is None:
        fontsize = 13
    
    assets_source_pct_value = assets_source_pct*100 

    # print(f'{assets_source_pct_value = }')
    # ensure the digits are readable
    if assets_source_pct_value >= 1:
        assets_source_pct_fmt = f'{assets_source_pct_value:.0f}%'
    elif assets_source_pct_value >= 0.1:
        assets_source_pct_fmt = f'{assets_source_pct_value:.1f}%'
    elif assets_source_pct_value >= 0.01:
        assets_source_pct_fmt = f'{(assets_source_pct_value*10):.1f}‰'
    else:
        assets_source_pct_fmt = f'<0.1‰'

    ax.text(0, 0, assets_source_pct_fmt, 
            fontsize=fontsize,
            horizontalalignment='center',
            verticalalignment='center')




def preprocessing_for_energy_meter_with_benchmarking(df_meta_with_value_building,
                                                     period_column="period",  
                                                     kwh_column="kwh"):

    df_period_total = df_meta_with_value_building.groupby([period_column])[kwh_column].sum().div(1000).tail(2)

    # for avoiding the case that we don't have data for the previous week
    if df_period_total.shape[0] < 2:
        consumption_mwh_cur = df_period_total.to_list()[0]
        consumption_mwh_pre = consumption_mwh_cur 
    else:
        consumption_mwh_pre, consumption_mwh_cur = df_period_total.to_list()

    period_freqstr = df_period_total.index[-1].freqstr

    if period_freqstr == "M":
        days_in_period = df_period_total.index[-1].daysinmonth
    else:
        days_in_period = 7

    return consumption_mwh_cur, consumption_mwh_pre, days_in_period




def validate_data_not_empty(df_meta_with_value, df_meta_with_value_building):
    
    if df_meta_with_value.shape[0] == 0:
        raise Exception(f'The dataframe [df_meta_with_value] is empty!')
    
    if df_meta_with_value_building.shape[0] == 0:
        raise Exception(f'The dataframe [df_meta_with_value_building] is empty!')



import pandas as pd
from tqdm import tqdm

from .load_iot_data_partition import load_iot_data_partition

def load_iot_data(engine, time_range, iot_table_name, id_list, chunksize=10):

    # time_partition_list = generate_time_partition_list(time_range, freq=partition_freq)

    df_iot = pd.DataFrame([])

    id_list_partitions = [id_list[x:x+chunksize] for x in range(0, len(id_list), chunksize)]

    no_partition = len(id_list_partitions)

    # for id_list_part in tqdm(id_list_partitions):
    for batch, id_list_part in enumerate(id_list_partitions):

        print(f'batch: {batch+1}/{no_partition}')
        df_iot_part = load_iot_data_partition(engine, time_range, iot_table_name, id_list_part)
        # df_iot_part = load_iot_data_partition(engine, time_partition, iot_table_name, meta_table_name, statement_full)

        df_iot = pd.concat([df_iot, df_iot_part])
        
    return df_iot

    # load_iot_data_partition(engine, time_partition, iot_table_name, meta_table_name, statement_full)




def create_table(engine, table_name, columns_dict, hypertable=False, time_column='time', index=None, constraint=None):
    """
    Function Name: create_table
    Parameters   : sql connection engine, table name, column dictionay, hypertable required flag
    Description  : creates a new table 
    
    """
    
    if hypertable:     
        if time_column not in columns_dict:
            raise "For hypertable requires a valid time column in columns_dict defined with schema timestamp"
    
    with engine.connect() as conn: 
        
        if hypertable:
            # enable the timescaledb extension, so function create_hypertable can be used
            query_create_extensions = "CREATE EXTENSION IF NOT EXISTS timescaledb;" 
            conn.execute(query_create_extensions)
        
        columns_list = [f"{column_name} {columns_dict[column_name]}" for column_name in columns_dict]
        
        query_columns_statement = ", ".join(columns_list)
        
        query_create_normaltable = f"""CREATE TABLE {table_name} ({query_columns_statement});"""

        conn.execute(query_create_normaltable)

        table_feature_statement = ''
        
        if hypertable:
            query_create_hypertable = f"SELECT create_hypertable('{table_name}', '{time_column}');"
            conn.execute(query_create_hypertable)
            table_feature_statement += ', as hypertable'

        if index is not None:
            index_name = list(index)[0]
            if isinstance(index[index_name], list):
                index_columns_list = index[index_name]
            else:
                index_columns_list = [index[index_name]] # convert variable to list

            index_columns_str = ", ".join(index_columns_list)

            query_create_index = f"CREATE INDEX {index_name} ON {table_name} ({index_columns_str});"
            conn.execute(query_create_index)
            table_feature_statement += f', with index: {index_name} - [{index_columns_str}]'

        if constraint is not None:
            for constraint_name in list(constraint):
                if isinstance(constraint[constraint_name], list):
                    constraint_columns_list = constraint[constraint_name]
                else:
                    constraint_columns_list = [constraint[constraint_name]] # convert variable to list

                constraint_columns_str = ", ".join(constraint_columns_list)

                query_create_constraint = f"ALTER TABLE {table_name} ADD CONSTRAINT {constraint_name} UNIQUE ({constraint_columns_str});"
                
                # print('query_create_constraint: ', query_create_constraint)
                conn.execute(query_create_constraint)

                table_feature_statement += f', with constraint: {constraint_name} - [{constraint_columns_str}]'
                # conn.commit()

                # example:
                # ALTER TABLE table_name
                # ADD CONSTRAINT constraint_name
                # UNIQUE(column1,column2,..)

    print(f'New table [{table_name}] is created' + table_feature_statement + '.')
    return



from get_conn_by_id import get_conn_by_id 

email_conn = get_conn_by_id('mindsett_reporting_email')

server = email_conn['host']
port = email_conn['port']
send_from = email_conn['send_from']
username = email_conn['login']
password = email_conn['password']





def get_formatted_arrow(pct_value):

    if pct_value > 0.005: 
            label_arrow_str = r'${\blacktriangle}$'

    elif pct_value < -0.005:
        label_arrow_str = r'$\:\!\triangledown\:\!$'

    else:
        label_arrow_str = r'$\!$--'
    
    label_arrow_pad = ' '
    label_arrow = label_arrow_pad + label_arrow_str

    return label_arrow



from .load_cf_from_db import load_cf_from_db
from .sync_with_sharepoint import sync_with_sharepoint

def fetch_sharepoint_info(engine, 
                          sharepoint, 
                          table_name_cache, 
                          table_columns_cache,
                          sync_source=None):

    if sync_source is None:
        sync_source = True

    # get contacts/configs from the db
    df_cfg_db = load_cf_from_db(engine, table_name_cache, table_columns_cache)

    # print(f'{df_cfg_db.info()=}')

    if sync_source:
        df_cfg = sync_with_sharepoint(engine, 
                                    df_cfg_db, 
                                    sharepoint, 
                                    table_name_cache)
    else:
        df_cfg = df_cfg_db

    return df_cfg


from typing import List

def _fetchPrismConf(self, board_id:int)->List[dict]:
    confColumnIds = self._searchConfColumnIds(board_id=board_id)
    groupIds = self._fetchGroups(board_id=board_id)
    ret = []
    for gId in groupIds:
        items = self._fetchPrismConfByGroup(board_id=board_id, groupId=gId, columns=confColumnIds)
        ret.extend(items)
    return ret


from pathlib import Path
import os
import ast

class ClientSite:

    from .get_schedule_records import get_schedule_records
    from .get_period_range import get_period_range
    from .get_start_time import get_start_time
    from .get_start_time_co2_barchart import get_start_time_co2_barchart
    from .get_end_time import get_end_time

    pct_level_tobe_others = 0.03
    pct_hide = 4 # hide the labels smaller then the value specified by pct_hide
    currency = 'GBP'
    conv_mwh_price = 190
    timezone = 'UTC'
    size_in_sqm = True

    # setup default backup schedule info
    weekend=7
    working_start_time="08:30:00"
    working_end_time="23:59:59"

    period_offset_previous = 2 # offset between the current period and the previous finished period
    period_freq = 'W' # M for monthly and W for weekly
    period_offset = 1 # offset between the current period and the most recent finished period
    period_count = 6 # number of period to show in the co2 barchart
    group_name_column_exchange = {} # for some special (usually big) groups, to disclose further details by using its name from another column, e.g. {"Sockets": "circuit_description"}
    group_name_modification = {} # replacement name to make the long and complex name more readable

    occupancy_available = False
    asset_group = "thing_category"
    fillna_value = "Undefined"
    exception = {'link_type': ['Supply', 'Load to DB', 'Load to Exclude']}
    insight_statements = None # this will decide whether to show the insights or not

    working_start_time="00:00:00" #Default start and end time, may not be used (it is needed for the functions) if the schedule information from the database
    working_end_time="23:59:00"
    # print('group_name_modification: ', group_name_modification) # debug
    reviewer_emails = ['x.yang@cloudfmgroup.com', 
                    #    's.goud@cloudfmgroup.com',
                        's.ruthven@cloudfmgroup.com', 
    #                     'r.patel@cloudfmgroup.com',
                        'd.attoe@cloudfmgroup.com'
                      ]
    # reviewer_emails = ['x.yang@cloudfmgroup.com', 
    #                    's.goud@cloudfmgroup.com', 
    #                    's.ruthven@cloudfmgroup.com', 
    #                    'r.patel@cloudfmgroup.com']
    # cloudfm_cc_emails = ['x.yang@cloudfmgroup.com', 
    #                     's.goud@cloudfmgroup.com', 
    #                     's.ruthven@cloudfmgroup.com', 
    #                     'r.patel@cloudfmgroup.com',
    #                     'd.attoe@cloudfmgroup.com']
    
    # 12 March 2024: [Removed] ["d.attoe@cloudfmgroup.com","s.ruthven@cloudfmgroup.com", "r.patel@cloudfmgroup.com", 's.goud@cloudfmgroup.com']

    def __init__(self, site_item):

        site_dict = site_item.squeeze(axis=0).to_dict()

        # print('site_item: ', site_item)
        # print('site_dict: ', site_dict)

        # file_path = Path(site_obj.__file__)

        

        self.site_name = site_dict['building_name'].replace("'","''")
        self.org_name = site_dict['org'].replace("'","''")

        # if site_dict['bld_name_abbr'] == '':
        #     self.bld_name_abbr = self.site_name
        # else:
        #     self.bld_name_abbr = site_dict['bld_name_abbr']

        print('\n' + self.org_name + ' - ' + self.site_name + ': ')

        if site_dict['group_name_column_exchange'] != '':
            self.group_name_column_exchange = ast.literal_eval(site_dict['group_name_column_exchange'])
        
        if site_dict['group_name_modification'] != '':
            self.group_name_modification = ast.literal_eval(site_dict['group_name_modification'])

        if site_dict['insight_statements'] != '':
            self.insight_statements = ast.literal_eval(site_dict['insight_statements'])

        # monday_config = ['site_dict']

        for attribute in site_dict:
            if attribute not in ['building_name', 
                                 'publish', 
                                 'testing', 
                                 'org', 
                                 'send_to', 
                                 'mailing_list',
                                 'group_name_column_exchange', 
                                 'group_name_modification',
                                 'insight_statements',
                                 'last_update']:
                
                if 'email' not in attribute:
                    setattr(self, attribute, site_dict[attribute])
                    # print('debug ClientSite attr: ', attribute, site_dict[attribute])
        
        if self.manager_name == '':
            self.manager_name = self.site_name + ' ' + self.org_name

        
        # for attribute in dir(site_obj):
        #     if not attribute.startswith('__'):
        #         setattr(self, attribute, getattr(site_obj, attribute))

        # self.manager_emails += self.cloudfm_cc_emails

if __name__ == '__main__':
    a = ClientSite()
    print(type(a))